#include "jsweb.h"
#include "global.h"
#include "Test.h"

#ifdef JSDEBUGGER
static JSDContext *_jsdc;
#ifdef JSDEBUGGER_JAVA_UI
static JSDJContext *_jsdjc;
#endif /* JSDEBUGGER_JAVA_UI */
#endif /* JSDEBUGGER */

bool gReportWarnings= false;

static bool hasShebang(FILE* file)
{
    char c1= fgetc(file);
    char c2= fgetc(file);

    if ('#'!=c1 || '!'!=c2) {
        rewind(file);
        return false;
    }
    return true;
}

static bool
processShebangFile(JSContext* context, const char* shebangFile)
{
    JSScript *script;
    jsval result;
    FILE *file;
    
    file = fopen(shebangFile, "r");
    if (!file) {
        JS_ReportError(context, "Can't open file: %s: %s", shebangFile,
                       strerror(errno));
        return false;
    }

    /* Support the UNIX #! shell hack; gobble the first line if it starts
     * with '#'.  TODO - this isn't quite compatible with sharp variables,
     * as a legal js program (using sharp variables) might start with '#'.
     * But that would require multi-character lookahead.
     */
    char ch;
    if (hasShebang(file)) {
        while (EOF!=(ch = fgetc(file))) {
            if (ch == '\n' || ch == '\r')
                break;
        }
        ungetc(ch, file);
    }
        
    JSObject* obj= JS_GetGlobalObject(context);
    script = JS_CompileFileHandle(context, obj, shebangFile, file);
    if (!script)
        return false;
    
    JS_ExecuteScript(context, obj, script, &result);
    JS_DestroyScript(context, script);
    return true;
}

static bool
processArgs(JSContext* context, JSObject* obj, int argc, char **argv)
{
    int i;
    JSObject* argsObj;

    /*
     * Create arguments early and define it to root it, so it's safe from any
     * GC calls nested below, and so it is available to -f <file> arguments.
     */
    argsObj = JS_NewArrayObject(context, 0, NULL);
    if (!argsObj)
        return false;
    if (!JS_DefineProperty(context, obj, "arguments", OBJECT_TO_JSVAL(argsObj),
                           NULL, NULL, 0))
        return false;

    for (i = 0; i < argc; i++) {
        JSString *str = JS_NewStringCopyZ(context, argv[i]);
        if (!str)
            return false;
        if (!JS_DefineElement(context, argsObj, i, STRING_TO_JSVAL(str),
                              NULL, NULL, JSPROP_ENUMERATE)) {
            return false;
        }
    }
    // JS_ToggleOptions(context, JSOPTION_STRICT);
    // JS_ToggleOptions(context, JSOPTION_XML);
    return true;
}

void
reportError(JSContext* context, const char* message, JSErrorReport* report)
{
    int i, j, k, n;
    char *prefix, *tmp;
    const char *ctmp;

    if (!report) {
        fprintf(stderr, "%s\n", message);
        return;
    }

    /* Conditionally ignore reported warnings. */
    if (JSREPORT_IS_WARNING(report->flags) && !gReportWarnings)
        return;

    prefix = NULL;
    if (report->filename)
        prefix = JS_smprintf("%s:", report->filename);
    if (report->lineno) {
        tmp = prefix;
        prefix = JS_smprintf("%s%u: ", tmp ? tmp : "", report->lineno);
        JS_free(context, tmp);
    }
    if (JSREPORT_IS_WARNING(report->flags)) {
        tmp = prefix;
        prefix = JS_smprintf("%s%swarning: ",
                             tmp ? tmp : "",
                             JSREPORT_IS_STRICT(report->flags) ? "strict " : "");
        JS_free(context, tmp);
    }

    /* embedded newlines -- argh! */
    while ((ctmp = strchr(message, '\n')) != 0) {
        ctmp++;
        if (prefix)
            fputs(prefix, stderr);
        fwrite(message, 1, ctmp - message, stderr);
        message = ctmp;
    }

    /* If there were no filename or lineno, the prefix might be empty */
    if (prefix)
        fputs(prefix, stderr);
    fputs(message, stderr);

    if (!report->linebuf) {
        fputc('\n', stderr);
        goto out;
    }

    /* report->linebuf usually ends with a newline. */
    n = strlen(report->linebuf);
    fprintf(stderr, ":\n%s%s%s%s",
            prefix,
            report->linebuf,
            (n > 0 && report->linebuf[n-1] == '\n') ? "" : "\n",
            prefix);
    n = PTRDIFF(report->tokenptr, report->linebuf, char);
    for (i = j = 0; i < n; i++) {
        if (report->linebuf[i] == '\t') {
            for (k = (j + 8) & ~7; j < k; j++) {
                fputc('.', stderr);
            }
            continue;
        }
        fputc('.', stderr);
        j++;
    }
    fputs("^\n", stderr);
out:
    // if (!JSREPORT_IS_WARNING(report->flags)) {
    //     if (report->errorNumber == JSMSG_OUT_OF_MEMORY) {
    //         gExitCode = EXITCODE_OUT_OF_MEMORY;
    //     } else {
    //         gExitCode = EXITCODE_RUNTIME_ERROR;
    //     }
    // }
    JS_free(context, prefix);
}

int
main(int argc, char **argv)
{
    setlocale(LC_ALL, "");

#ifdef XP_OS2
   /* these streams are normally line buffered on OS/2 and need a \n, *
    * so we need to unbuffer then to get a reasonable prompt          */
    setbuf(stdout,0);
    setbuf(stderr,0);
#endif

    argc--;
    argv++;

    //  No startup file? die
    if (!argc)
        return 0;
    
    const char* startupFile= argv[0];

    argc--;
    argv++;
    
    JSRuntime* runtime= JS_NewRuntime(64L * 1024L * 1024L);
    JSContext* context= JS_NewContext(runtime, 8192);

    JS_SetErrorReporter(context, reportError);

    {
        JSAutoRequest request(context);
    
        JSObject* global= setupGlobal(context, NULL);

        setupTestClass(context, global);
    
        JS_SetVersion(context, (JSVersion)180);

        if (!processArgs(context, global, argc, argv))
            return 1;
    
    #ifdef JSDEBUGGER
        /*
        * XXX A command line option to enable debugging (or not) would be good
        */
        _jsdc = JSD_DebuggerOnForUser(runtime, NULL, NULL);
        if (!_jsdc)
            return 1;
        JSD_JSContextInUse(_jsdc, context);
    #ifdef JSD_LOWLEVEL_SOURCE
        JS_SetSourceHandler(runtime, SendSourceToJSDebugger, _jsdc);
    #endif /* JSD_LOWLEVEL_SOURCE */
    #ifdef JSDEBUGGER_C_UI
        JSDB_InitDebugger(runtime, _jsdc, 0);
    #endif /* JSDEBUGGER_C_UI */
    #endif /* JSDEBUGGER */

        if (!processShebangFile(context, startupFile))
            return 0;
	
    #ifdef JSDEBUGGER
        if (_jsdc)
            JSD_DebuggerOff(_jsdc);
    #endif  /* JSDEBUGGER */
    }
    
    printf("*** shutting down!\n");
    JS_DestroyContext(context);
    JS_DestroyRuntime(runtime);
    JS_ShutDown();
    
    return 0;
}
